/*-
 * Copyright (C)2008 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2008 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: one.c,v 1.2 2008/03/15 15:32:31 babolo Exp $";
#endif /* not lint */

#include <sys/param.h>
#include <sys/types.h>
#include <net/ethernet.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <sys/socket.h>
#include <sysexits.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>

#define argtimeval timeval                       /* borrowed from argus (Carnegie Mellon University) */
#define arg_uint8   u_char
#define arg_int8    char
#define arg_uint16  u_short
#define arg_int16   short
#define arg_uint32  u_int
#define arg_int32   int                      /* END borrowed from argus (Carnegie Mellon University) */

#include <cons_def.h>
#include <cons_out.h>
#include "argus1tocopy.h"

u_int32_t
argus1_policer(argus1tocopy *opts, u_int32_t ip) {
    u_int32_t p;

    for (p = 0; opts->policy[p].mask != 0;) {
        ifBLIN_QV3(opts->flags) 
            fprintf( stderr, "IP %08X mask %08X value %08X\n"
                   , ip, opts->policy[p].mask & ip, opts->policy[p].value
                   );
        if  ((opts->policy[p].mask & ip) == opts->policy[p].value) p = opts->policy[p].eq;
          else p = opts->policy[p].neq;
    }
    return(p);
}

static void
sumtraf(argus1tocopy *opts) {
    size_t p1;
    int j, k;
    argus1_traf *t;
#   define ENT(A)   (opts->btraf[A])
#   define POINTER  (opts->ptraf)
#   define LENGTH   (opts->ltraf)
#   define PRINTOUT {                                                                                 \
        for (k = 0; k <= POINTER; k++) {                                                              \
            fprintf( stderr, "%2d%c %08X %012llX %10u %10u %llu\n"                                    \
                   , k, (k < POINTER) ? ' ' : '*'                                                     \
                   , ENT(k)->ip                                                                       \
                   , ENT(k)->mac                                                                      \
                   , ENT(k)->time                                                                     \
                   , ENT(k)->category                                                                 \
                   , ENT(k)->byt                                                                      \
                   );                                                                                 \
    }   }

    ifBLIN_QV3(opts->flags) PRINTOUT
    p1 = POINTER;
    POINTER++;
    for (j = 0; j + 1 < POINTER; j++) {
        if  (  (ENT(j)->ip == ENT(p1)->ip)
            && (ENT(j)->mac == ENT(p1)->mac)
            && (ENT(j)->time == ENT(p1)->time)
            && (ENT(j)->category == ENT(p1)->category)
            ) {
            ENT(p1)->byt += ENT(j)->byt;
            ifBLIN_QV3(opts->flags) fprintf(stderr, "%d -> %d\n", j, p1);
            t = ENT(j);
            bcopy( &ENT(j + 1)
                 , &ENT(j)
                 , (p1 - j - 1) * sizeof(argus1_traf *)
                 );
            POINTER--;
            ENT(POINTER - 1) = ENT(p1);
            ENT(p1) = t;
            ifBLIN_QV4(opts->flags) PRINTOUT
            break;
    }   }
#   undef PRINTOUT
#   undef LENGTH
#   undef POINTER
#   undef ENT
}

static void
sumaddr(argus1tocopy *opts) {
    size_t p1;
    int j, k;
    argus1_addr *t;
#define ENT(A)  (opts->baddr[A])
#define POINTER (opts->paddr)
#define LENGTH  (opts->laddr)
#   define PRINTOUT {                                                                                 \
        for (k = 0; k <= POINTER; k++) {                                                              \
            fprintf( stderr, "%2d%c %08X %012llX %10u %llu %llu\n"                                    \
                   , k, (k < POINTER) ? ' ' : '*'                                                     \
                   , ENT(k)->ip                                                                       \
                   , ENT(k)->mac                                                                      \
                   , ENT(k)->time                                                                     \
                   , ENT(k)->pkt                                                                      \
                   , ENT(k)->sent                                                                     \
                   );                                                                                 \
    }   }

    ifBLIN_QV3(opts->flags) PRINTOUT
    p1 = POINTER;
    POINTER++;
    for (j = 0; j + 1 < POINTER; j++) {
        if  (  (ENT(j)->ip == ENT(p1)->ip)
            && (ENT(j)->mac == ENT(p1)->mac)
            && (ENT(j)->time == ENT(p1)->time)
            ) {
            ENT(p1)->pkt += ENT(j)->pkt;
            ENT(p1)->sent += ENT(j)->sent;
            ifBLIN_QV3(opts->flags) fprintf(stderr, "%d -> %d\n", j, p1);
            t = ENT(j);
            bcopy( &ENT(j + 1)
                 , &ENT(j)
                 , (p1 - j - 1) * sizeof(argus1_addr *)
                 );
            POINTER--;
            ENT(POINTER - 1) = ENT(p1);
            ENT(p1) = t;
            ifBLIN_QV4(opts->flags) PRINTOUT
            break;
    }   }
#   undef PRINTOUT
#   undef LENGTH
#   undef POINTER
#   undef ENT
}

static void
sumscan(argus1tocopy *opts) {
    size_t p1;
    int j, k;
    argus1_scan *t;
#define ENT(A)  (opts->bscan[A])
#define POINTER (opts->pscan)
#define LENGTH  (opts->lscan)
#   define PRINTOUT {                                                                                 \
        for (k = 0; k <= POINTER; k++) {                                                              \
            fprintf( stderr, "%2d%c %08X %012llX %10u %llu\n"                                         \
                   , k, (k < POINTER) ? ' ' : '*'                                                     \
                   , ENT(k)->ip                                                                       \
                   , ENT(k)->mac                                                                      \
                   , ENT(k)->time                                                                     \
                   , ENT(k)->nrec                                                                     \
                   );                                                                                 \
    }   }

    ifBLIN_QV3(opts->flags) PRINTOUT
    p1 = POINTER;
    POINTER++;
    for (j = 0; j + 1 < POINTER; j++) {
        if  (  (ENT(j)->ip == ENT(p1)->ip)
            && (ENT(j)->mac == ENT(p1)->mac)
            && (ENT(j)->time == ENT(p1)->time)
            ) {
            ENT(p1)->nrec += ENT(j)->nrec;
            ifBLIN_QV3(opts->flags) fprintf(stderr, "%d -> %d\n", j, p1);
            t = ENT(j);
            bcopy( &ENT(j + 1)
                 , &ENT(j)
                 , (p1 - j - 1) * sizeof(argus1_scan *)
                 );
            POINTER--;
            ENT(POINTER - 1) = ENT(p1);
            ENT(p1) = t;
            ifBLIN_QV4(opts->flags) PRINTOUT
            break;
    }   }
#   undef PRINTOUT
#   undef LENGTH
#   undef POINTER
#   undef ENT
}

static u_int64_t
eatoq(struct ether_addr *ea) {
    int i;
    u_int64_t ll;
    for (i = 0, ll = 0; i < ETHER_ADDR_LEN; i++) {
        ll <<= 8;
        ll |= ea->octet[i];
    }
    return(ll);
}

static int
ok(argus1tocopy *opts, u_int64_t mac) {
    int i;

    for (i = 0; i < opts->nmacs; i++) {
        if  (mac == opts->maclist[i]) return(0);
    }
    return(1);
}

#define B0 (opts->buffer[opts->bufpnt])
#define B1 (opts->buffer[opts->bufpnt + 1])

int
argus1_decode(argus1tocopy *opts, void *n) {
    int ex = EX_OK;
    struct WriteStruct *next = n;
    u_int32_t status = ntohl(next->status);
    u_int64_t srccnt, dstcnt, srcby, dstby;

    if  (!next) {
        ifBLIN_QV1(opts->flags) warnx("No next argus record");
        errno = ENOENT;
        ex = EX_DATAERR;
        goto out;
    }
    if  (status & ARGUSCONTROL) {
        /* statistics ? XXXX */
    } else if (  status & IPPROTO
    /*        || status & ARPPROTO */  /* XXXXX */
              ) {
        u_int32_t srcpol, dstpol, srcip, dstip;
        u_int64_t srcmac, dstmac;
        long time = ntohl(next->ws_ip.startime.tv_sec) | opts->timask;

        srcmac = eatoq(&next->ws_ip_phys.ethersrc);
        dstmac = eatoq(&next->ws_ip_phys.etherdst);
        srcip = ntohl(next->ws_ip_src.s_addr);
        dstip = ntohl(next->ws_ip_dst.s_addr);
        srcpol = argus1_policer(opts, srcip);
        dstpol = argus1_policer(opts, dstip);
        srccnt = dstcnt = srcby = dstby = 0;

        if  (status & ICMPPROTO) {
            if  (status & CON_ESTABLISHED) {
                srccnt = ntohl(next->ws_ip_udp.src_count);
                dstcnt = ntohl(next->ws_ip_udp.dst_count);
            } else if (   (   (status & TIMED_OUT)
                          ?   ((unsigned char *)&next->ws_ip_udp.src_bytes)[0]
                          :   next->ws_ip_icmp.type
                          )
                      ==  ICMP_ECHOREPLY
                      ) {
                dstcnt = 1;
            } else {
                srccnt = 1;
            }
            srcby = opts->pluspacket * srccnt;
            dstby = opts->pluspacket * dstcnt;
        } else if (status & (UDPPROTO | TCPPROTO)) {
            srccnt = ntohl(next->ws_ip_udp.src_count);
            dstcnt = ntohl(next->ws_ip_udp.dst_count);
            srcby = dstby = 0;
            if  (status & (CON_ESTABLISHED | UDPPROTO) || !(status & (SAW_SYN | SAW_SYN_SENT))) {
                srcby = ntohl(next->ws_ip_udp.src_bytes);
                dstby = ntohl(next->ws_ip_udp.dst_bytes);
            }
            srcby += opts->pluspacket * srccnt;
            dstby += opts->pluspacket * dstcnt;
        } else if (status & FRAG_ONLY) {
            srccnt = 1;
            dstcnt = 0;
            srcby = opts->pluspacket * srccnt + ntohs(next->ws_ip_frag.currlen);
            dstby = 0;
        } else {
            srccnt = ntohl(next->ws_ip_udp.src_count);
            dstcnt = ntohl(next->ws_ip_udp.dst_count);
            srcby = ntohl(next->ws_ip_udp.src_bytes) + opts->pluspacket * srccnt;
            dstby = ntohl(next->ws_ip_udp.dst_bytes) + opts->pluspacket * dstcnt;
        }
        if  (status & REVERSE) {
            u_int64_t t;
            t = srccnt; srccnt = dstcnt; dstcnt = t;
            t = srcby; srcby = dstby; dstby = t;
        }
        if  (ok(opts, srcmac)) {
            if  (dstby) {
                opts->btraf[opts->ptraf]->ip = srcip;
                opts->btraf[opts->ptraf]->mac = srcmac;
                opts->btraf[opts->ptraf]->time = time;
                opts->btraf[opts->ptraf]->category = opts->policy[dstpol].eq;
                opts->btraf[opts->ptraf]->byt = dstby;
                sumtraf(opts);
            }
            if  (srcby) {
                opts->btraf[opts->ptraf]->ip = srcip;
                opts->btraf[opts->ptraf]->mac = srcmac;
                opts->btraf[opts->ptraf]->time = time;
                opts->btraf[opts->ptraf]->category = opts->policy[dstpol].neq;
                opts->btraf[opts->ptraf]->byt = srcby;
                sumtraf(opts);
            }
            if  (dstcnt || srccnt) {
                opts->baddr[opts->paddr]->ip = srcip;
                opts->baddr[opts->paddr]->mac = srcmac;
                opts->baddr[opts->paddr]->time = time;
                opts->baddr[opts->paddr]->pkt = dstcnt;
                opts->baddr[opts->paddr]->sent = srccnt;
                sumaddr(opts);
            }
            opts->bscan[opts->pscan]->ip = srcip;
            opts->bscan[opts->pscan]->mac = srcmac;
            opts->bscan[opts->pscan]->time = time;
            opts->bscan[opts->pscan]->nrec = 1;
            sumscan(opts);
        }
        if  (ok(opts, dstmac)) {
            if  (dstby) {
                opts->btraf[opts->ptraf]->ip = dstip;
                opts->btraf[opts->ptraf]->mac = dstmac;
                opts->btraf[opts->ptraf]->time = time;
                opts->btraf[opts->ptraf]->category = opts->policy[srcpol].neq;
                opts->btraf[opts->ptraf]->byt = dstby;
                sumtraf(opts);
            }
            if  (srcby) {
                opts->btraf[opts->ptraf]->ip = dstip;
                opts->btraf[opts->ptraf]->mac = dstmac;
                opts->btraf[opts->ptraf]->time = time;
                opts->btraf[opts->ptraf]->category = opts->policy[srcpol].eq;
                opts->btraf[opts->ptraf]->byt = srcby;
                sumtraf(opts);
            }
            if  (dstcnt || srccnt) {
                opts->baddr[opts->paddr]->ip = dstip;
                opts->baddr[opts->paddr]->mac = dstmac;
                opts->baddr[opts->paddr]->time = time;
                opts->baddr[opts->paddr]->pkt = srccnt;
                opts->baddr[opts->paddr]->sent = dstcnt;
                sumaddr(opts);
            }
            opts->bscan[opts->pscan]->ip = dstip;
            opts->bscan[opts->pscan]->mac = dstmac;
            opts->bscan[opts->pscan]->time = time;
            opts->bscan[opts->pscan]->nrec = 1;
            sumscan(opts);
    }   }
out:
    return(ex);
}

static int
argus1_macpr(FILE *pr, u_int64_t mac) {
    return(fprintf( pr, "%02X:%02X:%02X:%02X:%02X:%02X"
                  , (u_int32_t)(mac >> 40) & 0xFF, (u_int32_t)(mac >> 32) & 0xFF
                  , (u_int32_t)(mac >> 24) & 0xFF, (u_int32_t)(mac >> 16) & 0xFF
                  , (u_int32_t)(mac >> 8) & 0xFF, (u_int32_t)mac & 0xFF
          )       );
}

void
argus1_trafout(argus1tocopy *opts) {
    argus1_traf *t;
    int k;

    if  (opts->itraf) {
        fprintf( opts->itraf, "%d.%d.%d.%d\t"
               , (opts->btraf[0]->ip >> 24) & 0xFF, (opts->btraf[0]->ip >> 16) & 0xFF
               , (opts->btraf[0]->ip >> 8) & 0xFF, opts->btraf[0]->ip & 0xFF
               );
        argus1_macpr( opts->itraf, opts->btraf[0]->mac);
        fprintf( opts->itraf, "\t%u\t%u\t%llu\n"
               , opts->btraf[0]->time
               , opts->btraf[0]->category
               , opts->btraf[0]->byt
               );
    }
    t = opts->btraf[0];
    bcopy(&opts->btraf[1], &opts->btraf[0], --(opts->ptraf) * sizeof(argus1_traf *));
    opts->btraf[opts->ptraf - 1] = t;
}

void
argus1_addrout(argus1tocopy *opts) {
    argus1_addr *t;
    int k;

    if  (opts->iaddr) {
        fprintf( opts->iaddr, "%d.%d.%d.%d\t"
               , (opts->baddr[0]->ip >> 24) & 0xFF, (opts->baddr[0]->ip >> 16) & 0xFF
               , (opts->baddr[0]->ip >> 8) & 0xFF, opts->baddr[0]->ip & 0xFF
               );
        argus1_macpr( opts->iaddr, opts->baddr[0]->mac);
        fprintf( opts->iaddr, "\t%u\t%llu\t%llu\n"
               , opts->baddr[0]->time
               , opts->baddr[0]->sent
               , opts->baddr[0]->pkt
               );
    }
    t = opts->baddr[0];
    bcopy(&opts->baddr[1], &opts->baddr[0], --(opts->paddr) * sizeof(argus1_addr *));
    opts->baddr[opts->paddr - 1] = t;
}

void
argus1_scanout(argus1tocopy *opts) {
    argus1_scan *t;
    int k;

    if  (opts->iscan) {
        fprintf( opts->iscan, "%d.%d.%d.%d\t"
               , (opts->bscan[0]->ip >> 24) & 0xFF, (opts->bscan[0]->ip >> 16) & 0xFF
               , (opts->bscan[0]->ip >> 8) & 0xFF, opts->bscan[0]->ip & 0xFF
               );
        argus1_macpr( opts->iscan, opts->bscan[0]->mac);
        fprintf( opts->iscan, "\t%u\t%llu\n"
               , opts->bscan[0]->time
               , opts->bscan[0]->nrec
               );
    }
    t = opts->bscan[0];
    bcopy(&opts->bscan[1], &opts->bscan[0], --(opts->pscan) * sizeof(argus1_scan *));
    opts->bscan[opts->pscan - 1] = t;
}

void
argus1_polout(argus1tocopy *opts, int c) {
    fprintf(stderr, "%6u ", c);
    if  (opts->policy[c].mask) {
        fprintf( stderr, "%d.%d.%d.%d\t       %d.%d.%d.%d"
               , (opts->policy[0].value >> 24) & 0xFF
               , (opts->policy[0].value >> 16) & 0xFF
               , (opts->policy[0].value >> 8) & 0xFF
               , opts->policy[0].value & 0xFF
               , (opts->policy[0].mask >> 24) & 0xFF
               , (opts->policy[0].mask >> 16) & 0xFF
               , (opts->policy[0].mask >> 8) & 0xFF
               , opts->policy[0].mask & 0xFF
               );
        fprintf(stderr, "\t%6u= %6u<>\n", opts->policy[c].eq, opts->policy[c].neq);
        if  ((opts->policy[c].value & opts->policy[c].mask) != opts->policy[c].value) {
            fprintf(stderr, "***** %6u rule illegal\n", c);
            fprintf(stderr, "**********************\n");
        }
    } else {
        fprintf(stderr, " %6uIn, %6uOut\n", opts->policy[c].eq, opts->policy[c].neq);
}   }
